@@ -25,6 +25,7 @@ from typing import List
 from typing import Tuple
 from typing import Optional
 from typing import Any
+from typing import Callable
 import os
 import os.path as path
 import codecs
@@ -32,6 +33,7 @@ import unicodedata
 import sqlite3
 import time
 import re
+import gzip
 import logging
 import itb_util
 import hunspell_suggest
@@ -181,17 +183,9 @@ class TabSqliteDb:
                 'Connect to the database %(name)s.',
                 {'name': self.user_db_file})
             self.database = sqlite3.connect(self.user_db_file)
-            self.database.executescript('''
-                PRAGMA encoding = "UTF-8";
-                PRAGMA case_sensitive_like = true;
-                PRAGMA page_size = 4096;
-                PRAGMA cache_size = 20000;
-                PRAGMA temp_store = MEMORY;
-                PRAGMA journal_mode = WAL;
-                PRAGMA journal_size_limit = 1000000;
-                PRAGMA synchronous = NORMAL;
-                ATTACH DATABASE "%s" AS user_db;
-            ''' % self.user_db_file)
+            self.set_database_pragma_options()
+            self.database.executescript(
+                'ATTACH DATABASE "%s" AS user_db;' % self.user_db_file)
         except Exception:
             LOGGER.exception(
                 'Could not open the database %(name)s.',
@@ -213,17 +207,9 @@ class TabSqliteDb:
                 {'name': self.user_db_file})
             self.init_user_db()
             self.database = sqlite3.connect(self.user_db_file)
-            self.database.executescript('''
-                PRAGMA encoding = "UTF-8";
-                PRAGMA case_sensitive_like = true;
-                PRAGMA page_size = 4096;
-                PRAGMA cache_size = 20000;
-                PRAGMA temp_store = MEMORY;
-                PRAGMA journal_mode = WAL;
-                PRAGMA journal_size_limit = 1000000;
-                PRAGMA synchronous = NORMAL;
-                ATTACH DATABASE "%s" AS user_db;
-            ''' % self.user_db_file)
+            self.set_database_pragma_options()
+            self.database.executescript(
+                'ATTACH DATABASE "%s" AS user_db;' % self.user_db_file)
         self.create_tables()
         if self.old_phrases:
             sqlargs = []
@@ -248,19 +234,6 @@ class TabSqliteDb:
             self.database.commit()
             self.database.execute('PRAGMA wal_checkpoint;')
 
-        # do not call this always on intialization for the moment.
-        # It makes the already slow “python engine/main.py --xml”
-        # to list the engines even slower and may break the listing
-        # of the engines completely if there is a problem with
-        # optimizing the databases. Probably bring this back as an
-        # option later if the code in self.optimize_database() is
-        # improved to do anything useful.
-        #try:
-        #    self.optimize_database()
-        #except:
-        #    print "exception in optimize_database()"
-        #    traceback.print_exc ()
-
         # try create all hunspell-tables in user database
         self.create_indexes(commit=False)
         self.generate_userdb_desc()
@@ -278,15 +251,11 @@ class TabSqliteDb:
         '''
         if not input_phrase or not phrase:
             return
-        input_phrase = itb_util.remove_accents(input_phrase)
-        input_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
+        input_phrase = itb_util.remove_accents(input_phrase.lower())
         phrase = unicodedata.normalize(
             itb_util.NORMALIZATION_FORM_INTERNAL, phrase)
-        p_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, p_phrase)
-        pp_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, pp_phrase)
+        p_phrase = itb_util.remove_accents(p_phrase.lower())
+        pp_phrase = itb_util.remove_accents(pp_phrase.lower())
         sqlstr = '''
         UPDATE user_db.phrases
         SET user_freq = :user_freq, timestamp = :timestamp
@@ -348,15 +317,11 @@ class TabSqliteDb:
                 user_freq)
         if not input_phrase or not phrase:
             return
-        input_phrase = itb_util.remove_accents(input_phrase)
-        input_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
+        input_phrase = itb_util.remove_accents(input_phrase.lower())
         phrase = unicodedata.normalize(
             itb_util.NORMALIZATION_FORM_INTERNAL, phrase)
-        p_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, p_phrase)
-        pp_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, pp_phrase)
+        p_phrase = itb_util.remove_accents(p_phrase.lower())
+        pp_phrase = itb_util.remove_accents(pp_phrase.lower())
         select_sqlstr = '''
         SELECT * FROM user_db.phrases
         WHERE input_phrase = :input_phrase
@@ -393,32 +358,6 @@ class TabSqliteDb:
         except Exception:
             LOGGER.exception('Unexpected error adding phrase to database.')
 
-    def optimize_database(self) -> None:
-        '''
-        Optimize the database by copying the contents
-        to temporary tables and back.
-        '''
-        sqlstr = '''
-            CREATE TABLE tmp AS SELECT * FROM %(database)s.phrases;
-            DELETE FROM user_db.phrases;
-            INSERT INTO user_db.phrases SELECT * FROM tmp ORDER BY
-            input_phrase, user_freq DESC, id ASC;
-            DROP TABLE tmp;'''
-        self.database.executescript(sqlstr)
-        self.database.executescript("VACUUM;")
-        self.database.commit()
-
-    def drop_indexes(self) -> None:
-        '''Drop the index in database to reduce it's size'''
-        sqlstr = '''
-            DROP INDEX IF EXISTS user_db.phrases_index_p;
-            DROP INDEX IF EXISTS user_db.phrases_index_i;
-            VACUUM;
-            '''
-
-        self.database.executescript(sqlstr)
-        self.database.commit()
-
     def create_indexes(self, commit: bool = True) -> None:
         '''Create indexes for the database.'''
         sqlstr = '''
@@ -433,22 +372,37 @@ class TabSqliteDb:
 
     def best_candidates(
             self,
-            phrase_frequencies: Dict[str, int]) -> List[Tuple[str, int]]:
+            phrase_frequencies: Dict[str, float],
+            title=False) -> List[Tuple[str, float]]:
         '''Sorts the phrase_frequencies dictionary and returns the best
         candidates.
+
+        Should *not* change the phrase_frequencies dictionary!
         '''
-        return sorted(phrase_frequencies.items(),
-                      key=lambda x: (
-                          -1*x[1],   # user_freq descending
-                          len(x[0]), # len(phrase) ascending
-                          x[0]       # phrase alphabetical
-                      ))[:20]
+        candidates = sorted(phrase_frequencies.items(),
+                            key=lambda x: (
+                                -1*x[1],   # user_freq descending
+                                len(x[0]), # len(phrase) ascending
+                                x[0]       # phrase alphabetical
+                            ))[:20]
+        if not title:
+            return candidates
+        candidates_title = []
+        phrases_title = set()
+        for candidate in candidates:
+            phrase = candidate[0]
+            phrase_title = phrase[:1].title() + phrase[1:]
+            if phrase_title in phrases_title:
+                continue
+            candidates_title.append((phrase_title, candidate[1]))
+            phrases_title.add(phrase_title)
+        return candidates_title
 
     def select_words(
             self,
             input_phrase: str,
             p_phrase: str = '',
-            pp_phrase: str = '') -> List[Tuple[str, int]]:
+            pp_phrase: str = '') -> List[Tuple[str, float]]:
         '''
         Get phrases from database completing input_phrase.
 
@@ -458,17 +412,14 @@ class TabSqliteDb:
         '''
         input_phrase = unicodedata.normalize(
             itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
-        p_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, p_phrase)
-        pp_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, pp_phrase)
+        p_phrase = itb_util.remove_accents(p_phrase.lower())
+        pp_phrase = itb_util.remove_accents(pp_phrase.lower())
+        title_case = input_phrase.istitle()
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'input_phrase=%s p_phrase=%s pp_phrase=%s',
-                input_phrase.encode('UTF-8'),
-                p_phrase.encode('UTF-8'),
-                pp_phrase.encode('UTF-8'))
-        phrase_frequencies: Dict[str, int] = {}
+                input_phrase, p_phrase, pp_phrase)
+        phrase_frequencies: Dict[str, float] = {}
         if not ' ' in input_phrase:
             # Get suggestions from hunspell dictionaries. But only
             # if input_phrase does not contain spaces. The hunspell
@@ -481,7 +432,7 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'hunspell: best_candidates=%s',
-                self.best_candidates(phrase_frequencies))
+                self.best_candidates(phrase_frequencies, title=title_case))
         # Remove the accents *after* getting the hunspell candidates.
         # If the accents were removed before getting the hunspell candidates
         # an input phrase like “Glühwürmchen” would not be added as a
@@ -490,9 +441,7 @@ class TabSqliteDb:
         # is not in the German hunspell dictionary as a single word but
         # created by suffix and prefix rules, the accent insensitive match
         # in the German hunspell dictionary would not find it either.
-        input_phrase = itb_util.remove_accents(input_phrase)
-        input_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
+        input_phrase = itb_util.remove_accents(input_phrase.lower())
         # Now phrase_frequencies might contain something like this:
         #
         # {'code': 0, 'communicability': 0, 'cold': 0, 'colour': 0}
@@ -515,11 +464,14 @@ class TabSqliteDb:
         SELECT * FROM user_db.phrases
         WHERE input_phrase LIKE "%(quoted_input_phrase)s%%"
         ;''' % {'quoted_input_phrase': quoted_input_phrase}
-        self.database.execute(sqlstr)
+        try:
+            self.database.execute(sqlstr)
+        except Exception:
+            LOGGER.exception('Unexpected error in creating database view.')
         sqlargs = {'p_phrase': p_phrase, 'pp_phrase': pp_phrase}
         sqlstr = (
             'SELECT phrase, sum(user_freq) FROM like_input_phrase_view '
-            + 'GROUP BY phrase;')
+            'GROUP BY phrase;')
         try:
             # Get “unigram” data from user_db.
             #
@@ -547,7 +499,7 @@ class TabSqliteDb:
             # If no unigrams matched, bigrams and trigrams cannot
             # match either. We can stop here and return what we got
             # from hunspell.
-            return self.best_candidates(phrase_frequencies)
+            return self.best_candidates(phrase_frequencies, title=title_case)
         # Now normalize the unigram frequencies with the total count
         # (which is 11 in the above example), which gives us the
         # normalized result:
@@ -567,14 +519,14 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'Unigram best_candidates=%s',
-                self.best_candidates(phrase_frequencies))
+                self.best_candidates(phrase_frequencies, title=title_case))
         if not p_phrase:
             # If no context for bigram matching is available, return
             # what we have so far:
-            return self.best_candidates(phrase_frequencies)
+            return self.best_candidates(phrase_frequencies, title=title_case)
         sqlstr = (
             'SELECT phrase, sum(user_freq) FROM like_input_phrase_view '
-            + 'WHERE p_phrase = :p_phrase GROUP BY phrase;')
+            'WHERE p_phrase = :p_phrase GROUP BY phrase;')
         try:
             results_bi = self.database.execute(sqlstr, sqlargs).fetchall()
         except Exception:
@@ -582,11 +534,11 @@ class TabSqliteDb:
                 'Unexpected error getting “bigram” data from user_db')
         if not results_bi:
             # If no bigram could be matched, return what we have so far:
-            return self.best_candidates(phrase_frequencies)
+            return self.best_candidates(phrase_frequencies, title=title_case)
         # get the total count of p_phrase to normalize the bigram frequencies:
         sqlstr = (
             'SELECT sum(user_freq) FROM like_input_phrase_view '
-            + 'WHERE p_phrase = :p_phrase;')
+            'WHERE p_phrase = :p_phrase;')
         try:
             count_p_phrase = self.database.execute(
                 sqlstr, sqlargs).fetchall()[0][0]
@@ -604,14 +556,14 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'Bigram best_candidates=%s',
-                self.best_candidates(phrase_frequencies))
+                self.best_candidates(phrase_frequencies, title=title_case))
         if not pp_phrase:
             # If no context for trigram matching is available, return
             # what we have so far:
-            return self.best_candidates(phrase_frequencies)
+            return self.best_candidates(phrase_frequencies, title=title_case)
         sqlstr = ('SELECT phrase, sum(user_freq) FROM like_input_phrase_view '
-                  + 'WHERE p_phrase = :p_phrase '
-                  + 'AND pp_phrase = :pp_phrase GROUP BY phrase;')
+                  'WHERE p_phrase = :p_phrase '
+                  'AND pp_phrase = :pp_phrase GROUP BY phrase;')
         try:
             results_tri = self.database.execute(sqlstr, sqlargs).fetchall()
         except Exception:
@@ -619,12 +571,12 @@ class TabSqliteDb:
                 'Unexpected error getting “trigram” data from user_db')
         if not results_tri:
             # if no trigram could be matched, return what we have so far:
-            return self.best_candidates(phrase_frequencies)
+            return self.best_candidates(phrase_frequencies, title=title_case)
         # get the total count of (p_phrase, pp_phrase) pairs to
         # normalize the bigram frequencies:
         sqlstr = (
             'SELECT sum(user_freq) FROM like_input_phrase_view '
-            + 'WHERE p_phrase = :p_phrase AND pp_phrase = :pp_phrase;')
+            'WHERE p_phrase = :p_phrase AND pp_phrase = :pp_phrase;')
         try:
             count_pp_phrase_p_phrase = self.database.execute(
                 sqlstr, sqlargs).fetchall()[0][0]
@@ -644,8 +596,8 @@ class TabSqliteDb:
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'Trigram best_candidates=%s',
-                self.best_candidates(phrase_frequencies))
-        return self.best_candidates(phrase_frequencies)
+                self.best_candidates(phrase_frequencies, title=title_case))
+        return self.best_candidates(phrase_frequencies, title=title_case)
 
     def generate_userdb_desc(self) -> None:
         '''
@@ -675,22 +627,27 @@ class TabSqliteDb:
         if self.user_db_file == ':memory:':
             return
         if not path.exists(self.user_db_file):
-            database = sqlite3.connect(self.user_db_file)
-            # a database containing the complete German Hunspell
-            # dictionary has less then 6000 pages. 20000 pages
-            # should be enough to cache the complete database
-            # in most cases.
-            database.executescript('''
-                PRAGMA encoding = "UTF-8";
-                PRAGMA case_sensitive_like = true;
-                PRAGMA page_size = 4096;
-                PRAGMA cache_size = 20000;
-                PRAGMA temp_store = MEMORY;
-                PRAGMA journal_mode = WAL;
-                PRAGMA journal_size_limit = 1000000;
-                PRAGMA synchronous = NORMAL;
-            ''')
-            database.commit()
+            self.database = sqlite3.connect(self.user_db_file)
+            self.set_database_pragma_options()
+
+    def set_database_pragma_options(self) -> None:
+        '''Set options for the user database'''
+        # a database containing the complete German Hunspell
+        # dictionary has less then 6000 pages. 20000 pages
+        # should be enough to cache the complete database
+        # in most cases.
+        self.database.executescript('''
+        PRAGMA encoding = "UTF-8";
+        PRAGMA case_sensitive_like = true;
+        PRAGMA page_size = 4096;
+        PRAGMA cache_size = 20000;
+        PRAGMA temp_store = MEMORY;
+        PRAGMA journal_mode = WAL;
+        PRAGMA journal_size_limit = 1000000;
+        PRAGMA synchronous = NORMAL;
+        PRAGMA auto_vacuum = FULL;
+        ''')
+        self.database.commit()
 
     def get_database_desc(self, db_file: str) -> Optional[Dict[str, str]]:
         '''Get the description of the database'''
@@ -746,9 +703,6 @@ CREATE TABLE phrases (id INTEGER PRIMARY KEY, input_phrase TEXT, phrase TEXT, p_
 
     def list_user_shortcuts(self) -> List[Tuple[str, str]]:
         '''Returns a list of user defined shortcuts from the user database.
-
-        :rtype: List of tuples of strings: [(str, str), ...]
-
         '''
         sqlstr = '''
         SELECT input_phrase, phrase FROM user_db.phrases WHERE user_freq >= :freq
@@ -781,19 +735,14 @@ CREATE TABLE phrases (id INTEGER PRIMARY KEY, input_phrase TEXT, phrase TEXT, p_
             return
         phrase = unicodedata.normalize(
             itb_util.NORMALIZATION_FORM_INTERNAL, phrase)
-        p_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, p_phrase)
-        pp_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, pp_phrase)
-        input_phrase = itb_util.remove_accents(input_phrase)
-        input_phrase = unicodedata.normalize(
-            itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
+        p_phrase = itb_util.remove_accents(p_phrase.lower())
+        pp_phrase = itb_util.remove_accents(pp_phrase.lower())
+        input_phrase = itb_util.remove_accents(input_phrase.lower())
 
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 'phrase=%(p)s, input_phrase=%(t)s',
-                {'p': phrase.encode('UTF-8'),
-                 't': input_phrase.encode('UTF-8')})
+                {'p': phrase, 't': input_phrase})
 
         # There should never be more than 1 database row for the same
         # input_phrase *and* phrase. So the following query on
@@ -906,28 +855,55 @@ CREATE TABLE phrases (id INTEGER PRIMARY KEY, input_phrase TEXT, phrase TEXT, p_
         Read data to train the prediction from a text file.
 
         :param filename: Full path of the text file to read.
-        :type filename: String
         '''
         if not os.path.isfile(filename):
-            return False
+            filename += '.gz'
+            if not os.path.isfile(filename):
+                return False
+        open_function: Callable = open
+        if filename.endswith('.gz'):
+            open_function = gzip.open
         rows = self.database.execute(
             'SELECT input_phrase, phrase, p_phrase, pp_phrase, '
             + 'user_freq, timestamp FROM phrases;').fetchall()
+        rows = sorted(rows, key = lambda x: (x[5])) # sort by timestamp
+        time_min = time.time()
+        time_max = time.time()
+        if rows:
+            time_min = rows[0][5]
+            time_max = rows[-1][5]
+        # timestamp for added entries (timestamp of existing entries is kept):
+        time_new = time_min + 0.20 * (time_max - time_min)
+        LOGGER.info('Minimum timestamp in the database=%s',
+                    time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(time_min)))
+        LOGGER.info('Maximum timestamp in the database=%s',
+                    time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(time_max)))
+        LOGGER.info('New timestamp in the database=%s',
+                    time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(time_new)))
         p_token = ''
         pp_token = ''
         database_dict: Dict[Tuple[str, str, str, str], Dict[str, Any]] = {}
         for row in rows:
+            # itb_util.remove_accents() returns in NORMALIZATION_FORM_INTERNAL.
+            # row[1] (“phrase”) should already be in NORMALIZATION_FORM_INTERNAL
+            # but better convert it again here just to make sure.
+            input_phrase = itb_util.remove_accents(row[0].lower())
+            phrase = unicodedata.normalize(
+                itb_util.NORMALIZATION_FORM_INTERNAL, row[1])
+            p_phrase = itb_util.remove_accents(row[2].lower())
+            pp_phrase = itb_util.remove_accents(row[3].lower())
             database_dict.update([((row[0], row[1], row[2], row[3]),
-                                   {'input_phrase': row[0],
-                                    'phrase': row[1],
-                                    'p_phrase': row[2],
-                                    'pp_phrase': row[3],
+                                   {'input_phrase': input_phrase,
+                                    'phrase': phrase,
+                                    'p_phrase': p_phrase,
+                                    'pp_phrase': pp_phrase,
                                     'user_freq': row[4],
                                     'timestamp': row[5]}
                                   )])
         lines = []
         try:
-            with codecs.open(filename, encoding='UTF-8') as file_handle:
+            with open_function(
+                    filename, mode='rt', encoding='UTF-8') as file_handle:
                 lines = [
                     unicodedata.normalize(
                         itb_util.NORMALIZATION_FORM_INTERNAL, line)
@@ -938,17 +914,20 @@ CREATE TABLE phrases (id INTEGER PRIMARY KEY, input_phrase TEXT, phrase TEXT, p_
             return False
         for line in lines:
             for token in itb_util.tokenize(line):
-                key = (token, token, p_token, pp_token)
+                key = (itb_util.remove_accents(token.lower()),
+                       token,
+                       itb_util.remove_accents(p_token.lower()),
+                       itb_util.remove_accents(pp_token.lower()))
                 if key in database_dict:
                     database_dict[key]['user_freq'] += 1
-                    database_dict[key]['timestamp'] = time.time()
                 else:
-                    database_dict[key] = {'input_phrase': token,
-                                          'phrase': token,
-                                          'p_phrase': p_token,
-                                          'pp_phrase': pp_token,
-                                          'user_freq': 1,
-                                          'timestamp': time.time()}
+                    database_dict[key] = {
+                        'input_phrase': itb_util.remove_accents(token.lower()),
+                        'phrase': token,
+                        'p_phrase': itb_util.remove_accents(p_token.lower()),
+                        'pp_phrase': itb_util.remove_accents(pp_token.lower()),
+                        'user_freq': 1,
+                        'timestamp': time_new}
                 pp_token = p_token
                 p_token = token
         sqlargs = []
@@ -1003,3 +982,171 @@ CREATE TABLE phrases (id INTEGER PRIMARY KEY, input_phrase TEXT, phrase TEXT, p_
         except Exception:
             LOGGER.exception('Unexpected error dumping database.')
             return
+
+    def number_of_rows_in_database(self) -> int:
+        '''
+        Return the current number of rows in the database
+
+        (For debugging)
+        '''
+        try:
+            return len(self.database.execute(
+                "SELECT * FROM phrases;").fetchall())
+        except Exception:
+            LOGGER.exception('Unexpected error getting number of database rows.')
+            return -1
+
+    def cleanup_database(self, thread: bool = True) -> None:
+        '''Cleanup user database by expiring entries which have not been
+        used for a long time.
+
+        :param thread: Whether this function is called in a different thread or not.
+
+        Usually it is called in a thread started when Typing Booster
+        has just finished starting up and is ready for input.
+
+        But in unittest test cases it may be called with an in memory
+        database without starting a thread. In that case a new
+        database connection is not needed.
+
+        '''
+        if thread and self.user_db_file == ':memory:':
+            LOGGER.info('Database cleanup not needed for memory database.')
+            return
+        LOGGER.info('Database cleanup starting ...')
+        time_now = time.time()
+         # id, input_phrase, phrase, p_phrase, pp_phrase, user_freq, timestamp
+        rows: List[Tuple[int, str, str, str, str, int, float]] = []
+        try:
+            if thread:
+                # SQLite objects created in a thread can only be used in
+                # that same thread.  As the database cleanup is usually
+                # called in a separate thread, get a new connection:
+                database = sqlite3.connect(self.user_db_file)
+            else:
+                database = self.database
+            rows = database.execute("SELECT * FROM phrases;").fetchall()
+        except Exception:
+            LOGGER.exception('Exception when accessing database')
+            import traceback
+            traceback.print_exc()
+        if not rows:
+            return
+        rows = sorted(rows,
+                      key = lambda x: (
+                          x[5], # user_freq
+                          x[6], # timestamp
+                          x[0], # id
+                      ))
+        LOGGER.info('Total number of database rows to check=%s ', len(rows))
+        index = len(rows)
+        max_rows = 50000
+        number_delete_above_max = 0
+        rows_kept: List[Tuple[int, str, str, str, str, int, float]] = []
+        for row in rows:
+            user_freq = row[5]
+            if (index > max_rows
+                and user_freq < itb_util.SHORTCUT_USER_FREQ):
+                LOGGER.info('1st pass: deleting %s %s',
+                            repr(row),
+                            time.strftime("%Y-%m-%d %H:%M:%S",
+                                          time.gmtime(row[6])))
+                number_delete_above_max += 1
+                sqlstr_delete = 'DELETE from phrases WHERE id = :id;'
+                sqlargs_delete = {'id': row[0]}
+                try:
+                    database.execute(sqlstr_delete, sqlargs_delete)
+                except Exception:
+                    LOGGER.exception(
+                        '1st pass: exception deleting row from database.')
+                    import traceback
+                    traceback.print_exc()
+            else:
+                rows_kept.append(row)
+            index -= 1
+        LOGGER.info('1st pass: Number of rows deleted above maximum size='
+                    f'{number_delete_above_max}')
+        # As the first pass above removes rows sorted by count and
+        # then by timestamp, it will never remove rows with a higher
+        # count even if they are extremely old. Therefore, a second
+        # pass uses sorting only by timestamp in order to first decay and
+        # eventually remove some rows which have not been used for a
+        # long time as well, even if they have a higher count.
+        # In this second pass, the 0.1% oldest rows are checked
+        # and:
+        #
+        # - if user_freq == 1 remove the row
+        # - if user_freq > 1 divide user_freq by 2 and update timestamp to “now”
+        #
+        # 0.1% is really not much but I want to be careful not to remove
+        # too much when trying this out.
+        #
+        # sort kept rows by timestamp only instead of user_freq and timestamp:
+        rows_kept = sorted(rows_kept,
+                           key = lambda x: (
+                               x[6], # timestamp
+                               x[0], # id
+                           ))
+        index = len(rows_kept)
+        LOGGER.info('1st pass: Number of rows kept=%s', index)
+        index_decay = int(max_rows * 0.999)
+        LOGGER.info('2nd pass: Index for decay=%s', index_decay)
+        number_of_rows_to_decay = 0
+        number_of_rows_to_delete = 0
+        for row in rows_kept:
+            user_freq = row[5]
+            if (index > index_decay
+                and user_freq < itb_util.SHORTCUT_USER_FREQ):
+                if user_freq == 1:
+                    LOGGER.info('2nd pass: deleting %s %s',
+                                repr(row),
+                                time.strftime("%Y-%m-%d %H:%M:%S",
+                                              time.gmtime(row[6])))
+                    number_of_rows_to_delete += 1
+                    sqlstr_delete = 'DELETE from phrases WHERE id = :id;'
+                    sqlargs_delete = {'id': row[0]}
+                    try:
+                        database.execute(sqlstr_delete, sqlargs_delete)
+                    except Exception:
+                        LOGGER.exception(
+                            '2nd pass: exception deleting row from database.')
+                        import traceback
+                        traceback.print_exc()
+                else:
+                    LOGGER.info('2nd pass: decaying %s %s',
+                                repr(row),
+                                time.strftime("%Y-%m-%d %H:%M:%S",
+                                              time.gmtime(row[6])))
+                    number_of_rows_to_decay += 1
+                    sqlstr_update = '''
+                    UPDATE phrases
+                    SET user_freq = :user_freq, timestamp = :timestamp
+                    WHERE id = :id
+                    ;'''
+                    sqlargs_update = {'id': row[0],
+                                      'user_freq': int(user_freq/2),
+                                      'timestamp': time.time()}
+
+                    try:
+                        database.execute(sqlstr_update, sqlargs_update)
+                    except Exception:
+                        LOGGER.exception(
+                            '2nd pass: exception decaying row from database.')
+                        import traceback
+                        traceback.print_exc()
+            index -= 1
+        LOGGER.info('Commit database and execute checkpoint ...')
+        database.commit()
+        database.execute('PRAGMA wal_checkpoint;')
+        LOGGER.info('Rebuild database using VACUUM command ...')
+        database.execute('VACUUM;')
+        LOGGER.info('Number of database rows deleted=%s',
+                     number_delete_above_max + number_of_rows_to_delete)
+        LOGGER.info('Number of database rows decayed=%s',
+                    number_of_rows_to_decay)
+        LOGGER.info('Number of rows before cleanup=%s', len(rows))
+        LOGGER.info('Number of rows remaining=%s',
+                    len(rows_kept) - number_of_rows_to_delete)
+        LOGGER.info('Time for database cleanup=%s seconds',
+                    time.time() - time_now)
+        LOGGER.info('Database cleanup finished.')
